@@ -1,10 +1,11 @@
 # flake8: noqa
+from .blockmatch import *
 from .bm3d import *
 from .ccd import *
+from .dfttest import *
 from .funcs import *
 from .knlm import *
 from .mvtools import *
 from .postprocess import *
 from .prefilters import *
 from .regress import *
-from .wnnm import *
