@@ -10,7 +10,8 @@ from typing import TYPE_CHECKING, Any, Literal, cast, overload
 
 from vsaa import Nnedi3, Znedi3
 from vsexprtools import ExprOp, aka_expr_available, norm_expr
-from vskernels import Bicubic, BicubicZopti, Bilinear, KernelT, Scaler, ScalerT
+from vsmasktools import retinex
+from vskernels import Bicubic, BicubicZopti, Bilinear, Scaler, ScalerT
 from vsrgtools import bilateral, blur, gauss_blur, min_blur, replace_low_frequencies
 from vstools import (
     ColorRange, ConvMode, CustomEnum, CustomIntEnum, CustomRuntimeError, DitherType, PlanesT, SingleOrArr,
@@ -21,10 +22,11 @@ from vstools import (
 
 from .bm3d import BM3D as BM3DM
 from .bm3d import BM3DCPU, AbstractBM3D, BM3DCuda, BM3DCudaRTC, Profile
-from .knlm import DEVICETYPE, ChannelMode, DeviceType, knl_means_cl
+from .knlm import DEVICETYPE, DeviceType, nl_means
 
 __all__ = [
     'Prefilter', 'prefilter_to_full_range',
+    'MultiPrefilter',
     'PelType'
 ]
 
@@ -73,6 +75,20 @@ class PrefilterBase(CustomIntEnum, metaclass=PrefilterMeta):
             temp_thr, spat_thr = kwargs.get('temp_thr', 2), kwargs.get('spat_thr', 2)
             return min_blur(clip, 2, planes).flux.SmoothST(temp_thr, spat_thr, planes)
 
+        if pref_type == Prefilter.DFTTEST_SMOOTH:
+            sigma = kwargs.get('sigma', 128.0)
+            sigma2 = kwargs.get('sigma2', sigma / 16)
+            sbsize = kwargs.get('sbsize', 8 if clip.width > 1280 else 6)
+            sosize = kwargs.get('sosize', 6 if clip.width > 1280 else 4)
+
+            kwargs |= dict(
+                sbsize=sbsize, sosize=sosize, slocation=[
+                    0.0, sigma2, 0.05, sigma, 0.5, sigma, 0.75, sigma2, 1.0, 0.0
+                ]
+            )
+
+            pref_type = Prefilter.DFTTEST
+
         if pref_type == Prefilter.DFTTEST:
             dftt_args = dict[str, Any](tbsize=1, slocation=[0.0, 4.0, 0.2, 9.0, 1.0, 15.0]) | kwargs
 
@@ -87,14 +103,24 @@ class PrefilterBase(CustomIntEnum, metaclass=PrefilterMeta):
 
             return dfft.std.MaskedMerge(clip, pref_mask, planes)
 
-        if pref_type == Prefilter.KNLMEANSCL:
-            kwargs |= dict(strength=7.0, tr=1, sr=2, simr=2) | kwargs | dict(channels=ChannelMode.from_planes(planes))
-            knl = knl_means_cl(clip, **kwargs)
+        if pref_type == Prefilter.NLMEANS:
+            kwargs |= dict(strength=7.0, tr=1, sr=2, simr=2) | kwargs | dict(planes=planes)
+            knl = nl_means(clip, **kwargs)
 
             return replace_low_frequencies(knl, clip, 600 * (clip.width / 1920), False, planes)
 
         if pref_type == Prefilter.BM3D:
-            bm3d_arch: type[AbstractBM3D] = kwargs.pop('arch', BM3DCuda if kwargs.pop('gpu', False) else BM3DM)
+            bm3d_arch: type[AbstractBM3D] = kwargs.pop('arch', None)
+            gpu: bool | None = kwargs.pop('gpu', None)
+
+            if gpu is None:
+                gpu = hasattr(core, 'bm3dcuda')
+
+            if bm3d_arch is None:
+                if gpu:  # type: ignore
+                    bm3d_arch = BM3DCudaRTC if hasattr(core, 'bm3dcuda_rtc') else BM3DCuda
+                else:
+                    bm3d_arch = BM3DCPU if hasattr(core, 'bm3dcpu') else BM3DM
 
             if bm3d_arch is BM3DM:
                 sigma, profile = 10, Profile.FAST
@@ -111,7 +137,7 @@ class PrefilterBase(CustomIntEnum, metaclass=PrefilterMeta):
 
             bm3d_args = dict[str, Any](sigma=sigmas, radius=1, profile=profile) | kwargs
 
-            return bm3d_arch(clip, **bm3d_args).clip
+            return bm3d_arch(clip, **bm3d_args).final()
 
         if pref_type == Prefilter.SCALEDBLUR:
             scale = kwargs.pop('scale', 2)
@@ -169,8 +195,8 @@ class PrefilterBase(CustomIntEnum, metaclass=PrefilterMeta):
             max_len = max(len(otherS), len(otherR))
 
             if max_len:
-                otherS = list[float | list[float]](reversed(normalize_seq(otherS or baseS, max_len)))  # type: ignore
-                otherR = list[float | list[float]](reversed(normalize_seq(otherR or baseR, max_len)))  # type: ignore
+                otherS = list[float | list[float]](reversed(normalize_seq(otherS or baseS, max_len)))
+                otherR = list[float | list[float]](reversed(normalize_seq(otherR or baseR, max_len)))
 
                 for siS, siR in zip(otherS, otherR):
                     base, ref = ref or clip, bilateral(base, siS, siR, ref, **kwargs)
@@ -208,8 +234,14 @@ class Prefilter(PrefilterBase):
     DFTTEST = 4
     """Denoising in frequency domain with dfttest and an adaptive mask for retaining lineart."""
 
-    KNLMEANSCL = 5
-    """Denoising with KNLMeansCL, then postprocessed to remove low frequencies."""
+    DFTTEST_SMOOTH = 12
+    """Denoising like in DFTTEST but with high defaults for lower frequencies."""
+
+    NLMEANS = 5
+    """Denoising with NLMeans, then postprocessed to remove low frequencies."""
+
+    KNLMEANSCL = NLMEANS
+    """Deprecated, use NLMEANS instead."""
 
     BM3D = 6
     """Normal spatio-temporal denoising using BM3D."""
@@ -266,15 +298,22 @@ class Prefilter(PrefilterBase):
 
         @overload
         def __call__(  # type: ignore
-            self: Literal[Prefilter.KNLMEANSCL], clip: vs.VideoNode, /, planes: PlanesT = None,
+            self: Literal[Prefilter.DFTTEST_SMOOTH], clip: vs.VideoNode, /, planes: PlanesT = None, *,
+            sigma: float = 128.0, sigma2: float | None = None, sbsize: int | None = None, sosize: int | None = None,
+            **kwargs: Any
+        ) -> vs.VideoNode:
+            ...
+
+        @overload
+        def __call__(  # type: ignore
+            self: Literal[Prefilter.NLMEANS], clip: vs.VideoNode, /, planes: PlanesT = None,
             *, strength: SingleOrArr[float] = 7.0, tr: SingleOrArr[int] = 1, sr: SingleOrArr[int] = 2,
             simr: SingleOrArr[int] = 2, device_type: DEVICETYPE | DeviceType = DeviceType.AUTO, **kwargs: Any
         ) -> vs.VideoNode:
             """
-            Denoising with KNLMeansCL, then postprocessed to remove low frequencies.
+            Denoising with NLMeans, then postprocessed to remove low frequencies.
 
-            :param clip:            Clip to be preprocessed.
-            :param planes:          Planes to be preprocessed.
+            :param clip:            Source clip.
             :param strength:        Controls the strength of the filtering.\n
                                     Larger values will remove more noise.
             :param tr:              Temporal Radius. Temporal size = `(2 * tr + 1)`.\n
@@ -288,8 +327,9 @@ class Prefilter(PrefilterBase):
             :param simr:            Similarity Radius. Similarity neighbourhood size = `(2 * simr + 1) ** 2`.\n
                                     Sets the radius of the similarity neighbourhood window.\n
                                     The impact on performance is low, therefore it depends on the nature of the noise.
-            :param device_type:     Set the OpenCL device to use for processing.
-            :param kwargs:          Additional arguments to pass to knlmeansCL.
+            :param planes:          Set the clip planes to be processed.
+            :param device_type:     Set the device to use for processing. The fastest device will be used by default.
+            :param kwargs:          Additional arguments passed to the plugin.
 
             :return:                Denoised clip.
             """
@@ -299,8 +339,7 @@ class Prefilter(PrefilterBase):
             self: Literal[Prefilter.BM3D], clip: vs.VideoNode, /, planes: PlanesT = None,
             *, arch: type[AbstractBM3D] = ..., gpu: bool = False,
             sigma: SingleOrArr[float] = ..., radius: SingleOrArr[int] = 1,
-            profile: Profile = ..., ref: vs.VideoNode | None = None, refine: int = 1,
-            yuv2rgb: KernelT = Bicubic, rgb2yuv: KernelT = Bicubic
+            profile: Profile = ..., ref: vs.VideoNode | None = None, refine: int = 1
         ) -> vs.VideoNode:
             """
             Normal spatio-temporal denoising using BM3D.
@@ -315,8 +354,6 @@ class Prefilter(PrefilterBase):
                                 * 0 means basic estimate only.
                                 * 1 means basic estimate with one final estimate.
                                 * n means basic estimate refined with final estimate for n times.
-            :param yuv2rgb:     Kernel used for converting the clip from YUV to RGB.
-            :param rgb2yuv:     Kernel used for converting back the clip from RGB to YUV.
 
             :return:            Preprocessed clip.
             """
@@ -469,14 +506,22 @@ class Prefilter(PrefilterBase):
 
         @overload
         def __call__(  # type: ignore
-            self: Literal[Prefilter.KNLMEANSCL], *, planes: PlanesT = None,
+            self: Literal[Prefilter.DFTTEST_SMOOTH], *, planes: PlanesT = None,
+            sigma: float = 128.0, sigma2: float | None = None, sbsize: int | None = None, sosize: int | None = None,
+            **kwargs: Any
+        ) -> PrefilterPartial:
+            ...
+
+        @overload
+        def __call__(  # type: ignore
+            self: Literal[Prefilter.NLMEANS], *, planes: PlanesT = None,
             strength: SingleOrArr[float] = 7.0, tr: SingleOrArr[int] = 1, sr: SingleOrArr[int] = 2,
             simr: SingleOrArr[int] = 2, device_type: DEVICETYPE | DeviceType = DeviceType.AUTO, **kwargs: Any
         ) -> PrefilterPartial:
             """
-            Denoising with KNLMeansCL, then postprocessed to remove low frequencies.
+            Denoising with NLMeans, then postprocessed to remove low frequencies.
 
-            :param planes:          Planes to be preprocessed.
+            :param planes:          Set the clip planes to be processed.
             :param strength:        Controls the strength of the filtering.\n
                                     Larger values will remove more noise.
             :param tr:              Temporal Radius. Temporal size = `(2 * tr + 1)`.\n
@@ -490,8 +535,8 @@ class Prefilter(PrefilterBase):
             :param simr:            Similarity Radius. Similarity neighbourhood size = `(2 * simr + 1) ** 2`.\n
                                     Sets the radius of the similarity neighbourhood window.\n
                                     The impact on performance is low, therefore it depends on the nature of the noise.
-            :param device_type:     Set the OpenCL device to use for processing.
-            :param kwargs:          Additional arguments to pass to knlmeansCL.
+            :param device_type:     Set the device to use for processing. The fastest device will be used by default.
+            :param kwargs:          Additional arguments passed to the plugin.
 
             :return:                Partial Prefilter.
             """
@@ -501,8 +546,7 @@ class Prefilter(PrefilterBase):
             self: Literal[Prefilter.BM3D], *, planes: PlanesT = None,
             arch: type[AbstractBM3D] = ..., gpu: bool = False,
             sigma: SingleOrArr[float] = ..., radius: SingleOrArr[int] = 1,
-            profile: Profile = ..., ref: vs.VideoNode | None = None, refine: int = 1,
-            yuv2rgb: KernelT = Bicubic, rgb2yuv: KernelT = Bicubic
+            profile: Profile = ..., ref: vs.VideoNode | None = None, refine: int = 1
         ) -> PrefilterPartial:
             """
             Normal spatio-temporal denoising using BM3D.
@@ -516,8 +560,6 @@ class Prefilter(PrefilterBase):
                                 * 0 means basic estimate only.
                                 * 1 means basic estimate with one final estimate.
                                 * n means basic estimate refined with final estimate for n times.
-            :param yuv2rgb:     Kernel used for converting the clip from YUV to RGB.
-            :param rgb2yuv:     Kernel used for converting back the clip from RGB to YUV.
 
             :return:            Partial Prefilter.
             """
@@ -671,6 +713,17 @@ class PrefilterPartial(PrefBase):  # type: ignore
         return self.prefilter(clip, planes=fallback(planes, self.planes), **kwargs | self.kwargs)  # type: ignore
 
 
+class MultiPrefilter(PrefBase):  # type: ignore
+    def __init__(self, *prefilters: Prefilter) -> None:
+        self.prefilters = prefilters
+
+    def __call__(self, clip: vs.VideoNode, /, **kwargs: Any) -> vs.VideoNode:  # type: ignore
+        for pref in self.prefilters:
+            clip = pref(clip)
+
+        return clip
+
+
 def prefilter_to_full_range(pref: vs.VideoNode, range_conversion: float, planes: PlanesT = None) -> vs.VideoNode:
     """
     Convert a limited range clip to full range.\n
@@ -721,7 +774,7 @@ def prefilter_to_full_range(pref: vs.VideoNode, range_conversion: float, planes:
 
         pref_full = norm_expr(work_clip, (luma_expr, f'x {neutral} - 128 * 112 / {neutral} +'), planes)
     elif range_conversion > 0.0:
-        pref_full = work_clip.retinex.MSRCP(None, range_conversion, None, False, True)
+        pref_full = retinex(work_clip, upper_thr=range_conversion, fast=False)
     else:
         pref_full = depth(
             work_clip, bits, range_out=ColorRange.FULL, range_in=ColorRange.LIMITED, dither_type=DitherType.NONE
